import logging
from pprint import pprint, pformat
import time
from copy import deepcopy
#from collections import defaultdict

from pymongo import UpdateOne, InsertOne
from pymongo.errors import BulkWriteError


def bulk_write(db, item_type, ops, stat=None, retries=3, log_first_failop=True):
    """
    Tries to apply `ops` Update operations to `item_type`
    collection with `bulk_write` method.
    Gives up if `retries` retries failed.

    Args:
        db - pymongo collection object
        item_type - name of collection
        ops - list of operations like UpdateOne
        stat - instance of `ioweb.stat:Stat`
        retries - number of retries
    """
    if stat:
        stat.inc('bulk-write-%s' % item_type)
    for retry in range(retries):
        try:
            res = db[item_type].bulk_write(ops, ordered=False)
        except BulkWriteError as ex:
            # TODO: repeat only failed operations!!!
            # TODO: repeat only in case of DUP (11000) errors
            if retry == (retries - 1):
                if log_first_failop:
                    logging.error(
                        'First failed operation:\n%s' % (
                            pformat(ex.details['writeErrors'][0])
                        )
                    )
                raise
            else:
                if stat:
                    stat.inc('bulk-write-%s-retry' % item_type)
        else:
            if stat:
                stat.inc(
                    'bulk-write-%s-upsert' % item_type,
                    res.bulk_api_result['nUpserted']
                )
                stat.inc(
                    'bulk-write-%s-change' % item_type,
                    res.bulk_api_result['nModified']
                )
            break
    return res


class BulkWriter(object):
    def __init__(self, db, item_type, bulk_size=100, stat=None, retries=3):
        self.db = db
        self.item_type = item_type
        self.stat = stat
        self.retries = retries
        self.bulk_size = bulk_size
        self.ops = []

    def _write_ops(self):
        res = bulk_write(self.db, self.item_type, self.ops, self.stat)
        self.ops = []
        return res

    def update_one(self, *args, **kwargs):
        self.ops.append(UpdateOne(*args, **kwargs))
        if len(self.ops) >= self.bulk_size:
            return self._write_ops()
        else:
            return None

    def insert_one(self, *args, **kwargs):
        self.ops.append(InsertOne(*args, **kwargs))
        if len(self.ops) >= self.bulk_size:
            return self._write_ops()
        else:
            return None

    def flush(self):
        if len(self.ops):
            return self._write_ops()
        else:
            return None


def iterate_collection(
        db, item_type, query, sort_field, iter_chunk=1000,
        fields=None, infinite=False, limit=None
    ):
    """
    Iterate over `db[item_type]` collection items matching `query`
    sorted by `sort_field`.

    Intenally, it fetches chunk of `iter_chunk` items at once and
    iterates over it. Then fetch next chunk.
    """
    recent_id = None
    count = 0
    query = deepcopy(query) # avoid side effects
    if sort_field in query:
        logging.error(
            'Function `iterate_collection` received query'
            ' that contains a key same as `sort_field`.'
            ' That might goes unexpected and weird ways.'
        )
    while True:
        if recent_id:
            query[sort_field] = {'$gt': recent_id}
        else:
            if sort_field in query:
                del query[sort_field]
        items = list(db[item_type].find(
            query, fields, sort=[(sort_field, 1)], limit=iter_chunk
        ))
        if not items:
            if infinite:
                sleep_time = 5
                logging.debug(
                    'No items to process. Sleeping %d seconds'
                    % sleep_time
                )
                time.sleep(sleep_time)
                recent_id = None
            else:
                return
        else:
            for item in items:
                yield item
                recent_id = item[sort_field]
                count += 1
                if limit and count >= limit:
                    return


def bulk_dup_insert(db, item_type, ops, dup_key, stat=None):
    if stat:
        stat.inc('bulk-dup-insert-%s' % item_type, len(ops))
    all_keys = set()
    uniq_ops = []
    for op in ops:
        if not isinstance(op, InsertOne):
            raise Exception(
                'Function bulk_dup_insert accepts only'
                ' InsertOne operations. Got: %s'
                % op.__class__.__name__
            )
        if dup_key not in op._doc:
            raise Exception(
                'Operation for bulk_dup_insert'
                ' does not have key "%s": %s'
                % (dup_key, str(op._doc)[:1000])
            )
        key = op._doc[dup_key]
        if key not in all_keys:
            all_keys.add(key)
            uniq_ops.append(op)

    try:
        db[item_type].bulk_write(uniq_ops, ordered=False)
    except BulkWriteError as ex:
        if (
                all(x['code'] == 11000 for x in ex.details['writeErrors'])
                and
                not ex.details['writeConcernErrors']
            ):
            error_keys = set(
                x['op'][dup_key] for x in ex.details['writeErrors']
            )
            res_keys = list(all_keys - error_keys)
            if stat:
                stat.inc(
                    'bulk-dup-insert-%s-inserted' % item_type,
                    len(res_keys)
                )
            return res_keys
        else:
            raise
    else:
        if stat:
            stat.inc(
                'bulk-dup-insert-%s-inserted' % item_type,
                len(all_keys)
            )
        return list(all_keys)


def bulk_simple_insert(db, item_type, ops, stat=None):
    if stat:
        stat.inc('bulk-dup-insert-%s' % item_type, len(ops))
    for op in ops:
        if not isinstance(op, InsertOne):
            raise Exception(
                'Function simple_bulk_insert accepts only'
                ' InsertOne operations. Got: %s'
                % op.__class__.__name__
            )
    try:
        db[item_type].bulk_write(ops, ordered=False)
    except BulkWriteError as ex:
        if (
                all(x['code'] == 11000 for x in ex.details['writeErrors'])
                and
                not ex.details['writeConcernErrors']
            ):
            if stat:
                stat.inc(
                    'bulk-dup-insert-%s-inserted' % item_type,
                    len(ops) - len(ex.details['writeErrors'])
                )
        else:
            raise
    else:
        if stat:
            stat.inc(
                'bulk-dup-insert-%s-inserted' % item_type,
                len(ops)
            )



import json
from pprint import pprint, pformat
import logging
from threading import Thread
import time
import socket

from ioweb import request
from ioweb.error import NetworkError


class ControlThread(object):
    def __init__(self, bot, ping_url, token, tags=None):
        self.bot = bot
        self.ping_url = ping_url
        self.token = token
        self.tags = (tags or {})

    def start(self):
        th = Thread(target=self.thread_worker)
        th.daemon = True
        th.start()

    def thread_worker(self):
        op_time = 0
        op_period = 10
        while True:
            now = time.time()
            if now - op_time > op_period:
                try:
                    # pytype: disable=not-callable
                    res = request(
                        self.ping_url,
                        headers={'access-token': self.token},
                        data=json.dumps(self.build_ping_data()),
                        timeout=5
                    )
                    # pytype: enable=not-callable
                except NetworkError as ex:
                    logging.exception('Network error in control thread')
                else:
                    if res.status != 200 or res.text != 'OK':
                        logging.error(
                            'Control server return unexpected response on ping request:'
                            ' code=%s, data=%s'
                            % (res.status, res.text[:50])
                        )
                finally:
                    op_time = now

    def build_ping_data(self):
        # init
        net = self.bot.network
        ret = {
            'handlers': [],
            'num_active_handlers': len(net.active_handlers),
            'num_idle_handlers': len(net.idle_handlers),
            'taskq_size': self.bot.taskq.qsize(),
            'resultq_size': self.bot.resultq.qsize(),
            'tags': self.tags,
            'hostname': socket.gethostname(),
            'crawler_id': self.bot.__class__.__name__,
            'session_id': self.bot.uuid,
            'stat': dict(self.bot.stat.total_counters),
        }

        total_active_time = 0
        for hdl in net.active_handlers:
            now = time.time()
            req = net.registry[hdl]['request']
            start_time = net.registry[hdl]['start']
            item = {
                'start_time': start_time,
                'active_time': (now - start_time) if start_time else 0,
                'retry_count': req.retry_count if req else None,
                'url': req['url'] if req else None,
                # todo: all req data
            }
            ret['handlers'].append(item)
            total_active_time += item['active_time']

        ret['median_handler_active_time'] = (
            (total_active_time / len(ret['handlers']))
            if len(ret['handlers']) else 0
        )
        return ret



from queue import Empty
import time
import sys
import logging
from collections import deque
from threading import Thread, Event
from uuid import uuid4
import traceback
from contextlib import contextmanager

from urllib3 import PoolManager
# This is only part of ioweb where I use gevent
# explicitly. My idea is be able to run ioweb
# on native thread also.
# If ioweb is started in threaded mode then
# just use Dummy Timeout class
try:
    from gevent import Timeout
except ImportError:
    @contextmanager
    def Timeout(*args, **kwargs):
        yield


from .transport import Urllib3Transport
from .util import debug, Pause
from .response import Response
from .error import (
    DataNotValid, NetworkError, OperationTimeoutError, collect_error_context
)
from .request import Request, CallbackRequest

network_logger = logging.getLogger(__name__)


def log_network_request(req):
    if isinstance(req, Request):
        if req.retry_count > 0:
            retry_str = ' [retry=#%d]' % req.retry_count
        else:
            retry_str = ''
        if req['proxy']:
            proxy_str = ' [proxy=%s, type=%s, auth=%s]' % (
                req['proxy'],
                req['proxy_type'].upper(),
                ('YES' if req['proxy_auth'] else 'NO'),
            )
        else:
            proxy_str = ''
        network_logger.debug(
            '%s %s%s%s', req.method(), req['url'], retry_str, proxy_str
        )


class NetworkService(object):
    transport_class = Urllib3Transport

    def __init__(
            self,
            taskq,
            resultq,
            fatalq,
            resultq_size_limit=None,
            threads=3,
            shutdown_event=None,
            setup_request_hook=None,
            prepare_response_hook=None,
            setup_request_proxy_hook=None,
            stat=None,
        ):
        # Input arguments
        self.taskq = taskq
        self.resultq = resultq
        self.fatalq = fatalq
        self.num_threads = threads
        if resultq_size_limit is None:
            resultq_size_limit = self.num_threads * 2
        self.resultq_size_limit = resultq_size_limit
        self.shutdown_event = shutdown_event
        self.setup_request_hook = setup_request_hook
        self.prepare_response_hook = prepare_response_hook
        self.setup_request_proxy_hook = setup_request_proxy_hook
        self.setup_request_proxy_hook = setup_request_proxy_hook
        self.stat = stat
        # Init logic
        self.idle_handlers = set()
        self.active_handlers = set()
        self.registry = {}
        self.threads = {}

    def run(self):
        for _ in range(self.num_threads):
            ref = object()
            self.idle_handlers.add(ref)
            pause = Pause()
            transport = self.transport_class(
                prepare_response_hook=self.prepare_response_hook
            )
            self.registry[ref] = {
                'transport': transport,
                'request': None,
                'response': None,
                'start': None,
                'pause': pause,
            }
            th = Thread(
                target=self.thread_network,
                args=[ref, pause, transport],
            )
            th.daemon = True
            th.start()
            self.threads[ref] = th

    def thread_network(self, ref, pause, transport):
        req = None
        try:
            while not self.shutdown_event.is_set():
                if pause.pause_event.is_set():
                    pause.process_pause()
                    print('pause processed')

                req = None
                if self.resultq.qsize() < self.resultq_size_limit:
                    try:
                        prio, req = self.taskq.get(True, 0.1)
                    except Empty:
                        pass
                    if req:
                        self.idle_handlers.remove(ref)
                        self.active_handlers.add(ref)
                        try:
                            res = Response()
                            transport.prepare_request(req, res)
                            if self.setup_request_hook:
                                self.setup_request_hook(transport, req)
                            if self.setup_request_proxy_hook:
                                self.setup_request_proxy_hook(transport, req)
                                self.process_request(ref, transport, req, res)
                        finally:
                            req = None
                            self.free_handler(ref)
                else:
                    time.sleep(0.01)
        except Exception as ex:
            ctx = collect_error_context(req)
            self.fatalq.put((sys.exc_info(), ctx))


    def process_request(self, ref, transport, req, res):
        self.registry[ref]['request'] = req
        self.registry[ref]['response'] = res
        self.registry[ref]['start'] = time.time()
        log_network_request(req)
        try:
            timeout_time = req['timeout'] or 31536000
            with Timeout(
                    timeout_time,
                    OperationTimeoutError(
                        'Timed out while reading response',
                        Timeout(timeout_time),
                    )
                ):
                if isinstance(req, CallbackRequest):
                    req['network_callback'](req, res)
                else:
                    transport.request(req, res)
        except OperationTimeoutError as ex:
            #logging.error(ex)
            error = ex
        except (req.retry_errors or (NetworkError, DataNotValid)) as ex:
            #logging.error(ex)
            error = ex
        except Exception as ex:
            #logging.error(ex)
            raise
        else:
            error = None
        if isinstance(req, CallbackRequest):
            res.error = error
        else:
            transport.prepare_response(
                req, res, error, raise_network_error=False
            )
        self.resultq.put({
            'request': req,
            'response': res,
        })

    def free_handler(self, ref):
        self.active_handlers.remove(ref)
        self.idle_handlers.add(ref)
        self.registry[ref]['request'] = None
        self.registry[ref]['response'] = None
        self.registry[ref]['start'] = None



import time

from ioweb.crawler import Crawler, Request


class TestCrawler(Crawler):
    def task_generator(self):
        while True:
            yield Request(
                name='page',
                url='https://yandex.ru/robots.txt'
            )
            time.sleep(0.1)

    def handler_page(self, req, res):
        assert '# yandex.ru' in res.text

bot = TestCrawler()
bot.run()



from gevent.monkey import patch_all
patch_all()

try:
    import grpc
except ImportError:
    pass
else:
    if grpc.__version__ != '1.18.0':
        raise Exception('grpcio version must be 1.18.0')
    from grpc.experimental import gevent
    gevent.init_gevent()



import logging
from logging.handlers import RotatingFileHandler


def setup_fail_logging(
        filename, mode='a', max_bytes=(50 * 1024 * 1024),
        formatter='%(asctime)s [%(levelname)s] (%(name)s) %(message)s',
    ):
    logger = logging.getLogger()
    hdl = RotatingFileHandler(filename, mode, maxBytes=max_bytes)
    hdl.setFormatter(logging.Formatter(formatter))
    hdl.setLevel(logging.ERROR)
    logger.addHandler(hdl)



from ioweb.session import Session, request
from ioweb.crawler import Crawler, Request


def test_sess():
    sess = Session()
    res = sess.request('https://yandex.ru/robots.txt')
    assert '# yandex.ru' in res.text


def test_request():
    res = request('https://yandex.ru/robots.txt')
    assert '# yandex.ru' in res.text


def test_crawler():

    class TestCrawler(Crawler):
        def task_generator(self):
            yield Request(
                name='page',
                url='https://yandex.ru/robots.txt'
            )

            def test_page(self, req, res):
                assert '# yandex.ru' in res.text

    bot = TestCrawler()
    bot.run()
